/* $Id: mush.c,v 1.4 2003/04/10 08:52:44 mulix Exp $ */ 

#include <stdio.h> 
#include <unistd.h> 
#include <stdlib.h> 
#include <errno.h> 
#include <assert.h> 
#include <ctype.h> 
#include <limits.h> 

#include <sys/types.h> 
#include <sys/wait.h>

#include <readline/readline.h>
#include <readline/history.h>

#include "include/mush.h" 
#include "include/builtin.h" 

#include "util/util.h" 

static const char whitespace[] = "\n\t "; 

struct command_args* alloc_args(size_t max_size)
{
	char** args; 	
	size_t argssz; 
	struct command_args* a; 

	argssz = sizeof(*args) * max_size; 

	if (!(a = malloc(sizeof(*a))))
		die("oom"); 
	
	if (!(args = malloc(argssz)))
		die("oom\n"); 

	memset(args, 0, argssz); 

	a->max_size = max_size; 
	a->argv = args; 
	a->argc = 0; 

	return a; 
}

static void free_args(struct command_args* args)
{
	free(args->argv); 
	free(args); 
}

static void enlarge_args(struct command_args* args)
{
	size_t argssz = args->max_size * 2; 
	char** arr; 

	if (!(arr = malloc(argssz)))
		die("oom\n"); 

	memcpy(arr, args->argv, args->max_size * sizeof(*args->argv)); 
	/* args->argc isn't changed */ 
	args->max_size *= 2; 
}

struct command_args* tokenize_shell_command(char* inp)
{
	size_t num_args = 16; 
	struct command_args* args; 

	if (!(args = alloc_args(num_args)))
		die("oom\n"); 

	args->argc = 0; 
	do { 
		/* now go on until the whitespace ends */ 
		while (*inp && isspace(*inp))
			inp++; 

		/* if it's NULL, we're done */ 
		if (*inp == '\0')
			break; 

		/* else, it's a new arg! */ 
		args->argv[args->argc++] = inp; 

		if (args->argc == args->max_size)
			enlarge_args(args); 
		
		/* find where this token ends */ 
		while (*inp && !isspace(*inp))
			inp++; 

		/* if it's NULL, we're done */ 
		if (*inp == '\0')
			break; 

		*inp++ = '\0'; 

	} while (1); 

	args->argv[args->argc] = NULL; 
	
	return args; 
}

static inline int has_slash(const char* name)
{
	return (strchr(name, '/') != NULL); 
}

/* copied from dumpmem */ 
static char* find_in_path(struct command_args* args)
{
	char* exec = NULL;
	char* env = getenv("PATH"); 
	char* beg, *end; 
	char buf[PATH_MAX + 256]; 
	char* name = args->argv[0]; 

	if (!env)
		return NULL; 

	if (!(env = strdup(env)))
		return NULL; 

	beg = env; 

	/* check if name references an executable relative to cwd */ 
	if (has_slash(name) && !access(name, X_OK)) {
		exec = strdup(name); 
		goto free_env; 
	} 

	/* find the first : character. */
	while (1) {
		end = strchr(beg, ':'); 
		if (end)
			*end = '\0'; 

		(void)snprintf(buf, sizeof(buf), "%s/%s", beg, args->argv[0]); 

		if (!access(buf, X_OK)) {
			exec = strdup(buf); 
			break; 
		}
		if (!end)
			break; 
		++end; 
		beg = end; 
	}

 free_env: 
	free(env); 

	return exec; 
}

static void execute_and_wait(const struct command_args* args)
{
	pid_t son; 
	int ret; 
	int status; 

	son = fork(); 
	
	if (son == -1)
		die("fork: %s\n", strerror(errno)); 

	if (son == 0) { 
		if ((ret = execv(args->argv[0], args->argv))) 
			die("execv('%s'): %s\n", args->argv[0], 
			    strerror(errno)); 
	} 

	do { 
		ret = waitpid(son, &status, 0); 
		if (WIFEXITED(status)) 
			break; 

	} while(1); 
}

int shell_loop(int argc, char* argv[]) 
{
	char* inp; 
	struct command_args* args; 
	char* exec;
	const struct builtin* handler; 

	while ((inp = readline(MUSH_PROMPT))) { 

		if (!*inp)
			continue; 

		/* add the input line to the history */ 
		add_history(inp); 

		/* spread it into tokens */ 
		args = tokenize_shell_command(inp); 

		if ((handler = builtin_handler(args->argv[0]))) { 
			handler->func(args->argv[0], args); 
			goto next; 
		} 

		if (!(exec = find_in_path(args))) { 
			warn("could not execute '%s': no such executable\n", 
			     args->argv[0]); 
			goto next; 
		} 
		args->argv[0] = exec; 

		execute_and_wait(args); 
		free(exec); 

	next: 
		free(inp); 
		free_args(args); 
	}
	return 0; 
} 
